"""Random values workflow plugin module

Remove this and other example files after bootstrapping your project.
"""

from ast import literal_eval
from collections.abc import Sequence
from typing import Any

from cmem_plugin_base.dataintegration.context import ExecutionContext, ExecutionReport
from cmem_plugin_base.dataintegration.description import Icon, Plugin, PluginParameter
from cmem_plugin_base.dataintegration.entity import Entities, Entity, EntityPath
from cmem_plugin_base.dataintegration.parameter.password import Password, PasswordParameterType
from cmem_plugin_base.dataintegration.plugins import WorkflowPlugin
from cmem_plugin_base.dataintegration.ports import (
    FixedNumberOfInputs,
    UnknownSchemaPort,
)
from cmem_plugin_base.dataintegration.types import IntParameterType
from langchain_postgres import PGVector


class DataContainer:
    """Encapsulate the data to be added to the database."""

    def __init__(self):
        self.texts = []
        self.embeddings = []
        self.metadatas = []

    def add(self, text: str, embedding: list[float], metadata: dict) -> None:
        """Add objects to the respective lists."""
        self.texts.append(text)
        self.embeddings.append(embedding)
        self.metadatas.append(metadata)

    def clear(self) -> None:
        """Clear all three lists."""
        self.texts.clear()
        self.embeddings.clear()
        self.metadatas.clear()

    def size(self) -> int:
        """Return the size of the lists (assuming all lists have the same length)."""
        return len(self.texts)


@Plugin(
    label="Postgres Vector Store",
    description="Store embeddings into Postgres Vector Store.",
    documentation="""
This plugin workflow store embeddings into Postgres Vector Store.

The vector embeddings and its respective metadata are going to be stored into a collection inside
the Postgres Vector Store.
It is possible to specify either the name of the attributes containing the vectors as well as the
metadata.
""",
    icon=Icon(package=__package__, file_name="postgresql.svg"),
    parameters=[
        PluginParameter(
            name="host",
            label="Database Host",
            description="The hostname of the postgres database service.",
            default_value="pgvector",
        ),
        PluginParameter(
            name="port",
            label="Database Port",
            param_type=IntParameterType(),
            description="The port number of the postgres database service.",
            default_value=5432,
        ),
        PluginParameter(
            name="user",
            label="Database User",
            description="The account name used to login to the postgres database service.",
            default_value="pgvector",
        ),
        PluginParameter(
            name="password",
            label="Database Password",
            param_type=PasswordParameterType(),
            description="The password of the database account.",
        ),
        PluginParameter(
            name="database",
            label="Database Name",
            description="The database name.",
            default_value="pgvector",
        ),
        PluginParameter(
            name="collection_name",
            label="Collection Name",
            description="The name of the collection, where the embeddings are going to be stored.",
            default_value="my_collection",
        ),
        PluginParameter(
            name="pre_delete_collection",
            label="Pre Delete Collection",
            description="If set to true, then the collection will removed at the beginning.",
            default_value=True,
        ),
        PluginParameter(
            name="source_path",
            label="Source Path",
            description="The name of the path to use for reading the embedding source.",
            default_value="_embedding_source",
            advanced=True,
        ),
        PluginParameter(
            name="embedding_path",
            label="Embedding Path",
            description="The name of the path to use for reading the embeddings.",
            default_value="_embedding",
            advanced=True,
        ),
        PluginParameter(
            name="metadata_paths",
            label="Metadata Paths",
            description="The comma separated list path names to be used as metadata. "
            "Empty name means all paths "
            "(except embedding source and embedding) will be used",
            default_value="",
            advanced=True,
        ),
        PluginParameter(
            name="batch_processing_size",
            label="Batch Processing Size",
            description="The number of entries to be processed in batch.",
            default_value=100,
            advanced=True,
        ),
    ],
)
class PGVectorStorePlugin(WorkflowPlugin):
    """PGVectorStorePlugin: Enable the storage of vectors into Postgres Vector Store."""

    connection_string: str
    user: str
    password: str
    host: str
    port: int
    database: str
    collection_name: str
    source_path: str
    embedding_path: str
    metadata_paths: str
    batch_processing_size: int
    inputs: Sequence[Entities]
    db: PGVector
    report: ExecutionReport

    def __init__(  # noqa: PLR0913
        self,
        host: str = "pgvector",
        port: int = 5432,
        user: str = "pgvector",
        password: Password | str = "",
        database: str = "pgvector",
        collection_name: str = "my_collection",
        pre_delete_collection: bool = True,
        source_path: str = "_embedding_source",
        embedding_path: str = "_embedding",
        metadata_paths: str = "metadata",
        batch_processing_size: int = 1000,
    ) -> None:
        self.batch_processing_size = batch_processing_size
        self.collection_name = collection_name
        self.user = user
        self.host = host
        self.port = port
        self.database = database
        self.embedding_path = embedding_path
        self.metadata_paths = metadata_paths
        self.source_path = source_path

        self.output_port = None
        self.input_ports = FixedNumberOfInputs([UnknownSchemaPort()])
        str_password = self.password = password if isinstance(password, str) else password.decrypt()
        self.connection_string = (
            f"postgresql+psycopg://{user}:{str_password}@{host}:{port}/{database}"
        )

        self.report = ExecutionReport()
        self.report.operation = "store"
        self.report.operation_desc = "vectors stored"

        self.db = PGVector(
            collection_name=self.collection_name,
            connection=self.connection_string,
            embeddings=None,  # type: ignore  # noqa: PGH003
            use_jsonb=True,
            pre_delete_collection=pre_delete_collection,
        )

    def _update_report(self, count: int) -> None:
        self.report.entity_count = count
        self.execution_context.report.update(self.report)

    def _paths_to_metadata(
        self, paths: list[str], entity: Entity, schema_paths: list[EntityPath]
    ) -> dict[str, Any]:
        metadata_paths = [path for path in schema_paths if path.path in paths]
        return self._metadata(entity, metadata_paths)

    def _index_of(self, path_name: str, paths: list[EntityPath]) -> int:
        for path in paths:
            if path.path == path_name:
                return paths.index(path)
        return -1

    def _metadata(self, entity: Entity, schema_paths: list[EntityPath]) -> dict[str, Any]:
        entity_dict: dict[str, Any] = {}
        for path, values in zip(schema_paths, entity.values, strict=False):
            if path.path not in (self.embedding_path, self.source_path):
                entity_dict[path.path] = list(values)
        return entity_dict

    def _cancel_workflow(self) -> bool:
        """Cancel workflow"""
        try:
            if self.execution_context.workflow.status() == "Canceling":
                self.log.info("End task (Cancelled Workflow).")
                return True
        except AttributeError:
            pass
        return False

    def _process_entities(self, entities: Entities) -> None:
        schema_paths: list[EntityPath] = list(entities.schema.paths)
        text_path_idx: int = self._index_of(self.source_path, schema_paths)
        embedding_path_idx: int = self._index_of(self.embedding_path, schema_paths)

        if text_path_idx == -1 or embedding_path_idx == -1:
            raise ValueError("The data does not have a text or embedding paths defined.")

        n_processed_entries: int = 0
        container: DataContainer = DataContainer()
        self._update_report(n_processed_entries)
        for entity in entities.entities:
            container.add(
                entity.values[text_path_idx][0],
                literal_eval(entity.values[embedding_path_idx][0]),
                self._paths_to_metadata(self.metadata_paths.split(","), entity, schema_paths)
                if self.metadata_paths
                else self._metadata(entity, schema_paths),
            )
            if container.size() == self.batch_processing_size:
                self.db.add_embeddings(container.texts, container.embeddings, container.metadatas)
                n_processed_entries += container.size()
                self._update_report(n_processed_entries)
                container.clear()
            if self._cancel_workflow():
                return
        if container.size() > 0:
            self.db.add_embeddings(container.texts, container.embeddings, container.metadatas)
            n_processed_entries += container.size()
            self._update_report(n_processed_entries)

    def execute(
        self,
        inputs: Sequence[Entities],
        context: ExecutionContext,
    ) -> None:
        """Run the workflow operator."""
        self.log.info("Start storing vectors.")
        self.inputs = inputs
        self.execution_context = context
        try:
            first_input: Entities = self.inputs[0]
        except IndexError as error:
            raise ValueError("Input port not connected.") from error
        self._process_entities(first_input)
        self.log.info("Vectors stored successfuly.")
